class MetaInfoField:
    def __init__(self, name: str, data_type: str, size: int, scale: int):
        self.name = name
        self.data_type = data_type
        self.size = size
        self.scale = scale
